#!/usr/bin/python3.8

"""
Date Created:
Date Edited: Tuesday 31 May 2022
Author: Mitchell T Dennis

A script to create neural networks which classify input data. The input data is of the form Mass, Helium Abundance, and Metallicity. The two networks are learning for good vs bad stellar models and good vs too old vs shell flashing models respectively.

"""
import tensorflow as tf
import numpy as np
import pandas as pd
import datetime
from pathlib import Path
import sys
import matplotlib.pyplot as plt
from tensorflow.keras import metrics
from tensorflow.keras.callbacks import EarlyStopping, LearningRateScheduler, ReduceLROnPlateau
from machineLearningFunctions import *
from sklearn.metrics import roc_curve

# Set initial constants
hiddenNodes = 256
scale = 1
layers = 10
dropout = 0
numRuns = int(sys.argv[1])

now = datetime.datetime.now()
dt_string = now.strftime("%d-%m-%y_%H-%M-%S")
home = str(Path.home())

# Open data files
inputFilePath = '/home/mtde226/MESA_emulator/LOGS/LOGS_'+str(numRuns)+'/postProcessResults/fullComboData.txt'
outputFilePath = '/home/mtde226/MESA_emulator/LOGS/LOGS_'+str(numRuns)+'/MachineLearningResults/'
normFile = open(outputFilePath+'normalisationConstantsCombo.txt', 'r')
# TwoFlagsLoss = open(outputFilePath+'2FlagsLossStatistics.txt', 'w')
# ThreeFlagsLoss = open(outputFilePath+'3FlagsLossStatistics.txt', 'w')

# Create dataframes and clean data
data = pd.read_csv(inputFilePath, header=0, index_col=False)
inputCols = ['mass', 'Y', 'Z', 'A']
outputCols1 = ['flag']
outputCols2 = ['flag0', 'flag2', 'flag3']

dataCols1 = inputCols+outputCols1
dataCols2 = inputCols+outputCols2

# Read the normalisation constants
massABSMin, massMin, massMax = [float(x) for x in normFile.readline().split()]
heliumABSMin, heliumMin, heliumMax = [float(x) for x in normFile.readline().split()]
metallicityABSMin, metallicityMin, metallicityMax = [float(x) for x in normFile.readline().split()]
axionABSMin, axionMin, axionMax = [float(x) for x in normFile.readline().split()]
timeABSMin, timeMin, timeMax = [float(x) for x in normFile.readline().split()]
gravABSMin, gravMin, gravMax = [float(x) for x in normFile.readline().split()]
TeffABSMin, TeffMin, TeffMax = [float(x) for x in normFile.readline().split()]
FeHABSMin, FeHMin, FeHMax = [float(x) for x in normFile.readline().split()]
logLABSMin, logLMin, logLMax = [float(x) for x in normFile.readline().split()]
IBandABSMin, IBandMin, IBandMax = [float(x) for x in normFile.readline().split()]
IerrABSMin, IerrMin, IerrMax = [float(x) for x in normFile.readline().split()]

data.query('flag != 1', inplace=True)

# Always use deep copies of data to avoid interference between data sets. Transform data set 1 into a binary classification problem
data1 = data[dataCols1].copy(deep=True)
data1.replace(to_replace={'flag': 2}, value={'flag':1}, inplace=True)
data1.replace(to_replace={'flag': 3}, value={'flag':1}, inplace=True)
data2 = data[dataCols1].copy(deep=True)

# Create training and testing data
xTrain1, yTrain1, xTest1, yTest1, xVerify1, yVerify1 = splitData(data1, inputCols, outputCols1, 0.8, smote=True)
xTrain2, yTrain2, xTest2, yTest2, xVerify2, yVerify2 = splitData(data2, inputCols, outputCols1, 0.8, smote=True)

# multi-class classification works best when using dummy variables (one for each different class)
yTrain2 = pd.get_dummies(yTrain2, columns=['flag'])
yTest2 = pd.get_dummies(yTest2, columns=['flag'])
yVerify2 = pd.get_dummies(yVerify2, columns=['flag'])

# yTrain2.rename(columns={'flag_0.0': 'flag0', 'flag_2.0': 'flag2', 'flag_3.0': 'flag3'}, inplace=True)
# yTest2.rename(columns={'flag_0.0': 'flag0', 'flag_2.0': 'flag2', 'flag_3.0': 'flag3'}, inplace=True)
# yVerify2.rename(columns={'flag_0.0': 'flag0', 'flag_2.0': 'flag2', 'flag_3.0': 'flag3'}, inplace=True)

# Create the two models; one for TRGB atmospheric data and the other for IBand data
model1=createModel(inputDim=len(inputCols), outputDim=len(outputCols1), numLayers=layers, hiddenInit=hiddenNodes, dropoutFrac=dropout, scaling=scale)
model2=createModel(inputDim=len(inputCols), outputDim=len(outputCols2), numLayers=layers, hiddenInit=hiddenNodes, dropoutFrac=dropout, scaling=scale)

# IMPORTANT: Each network needs an independent optimizers, early stopping, plateau monitors, etc...
# Set the optimizers
optimizer1 = tf.keras.optimizers.Adam(learning_rate=0.004, beta_1=0.9, beta_2=0.9)
optimizer2 = tf.keras.optimizers.Adam(learning_rate=0.004, beta_1=0.9, beta_2=0.9)

# Set the loss functions
loss1=tf.keras.losses.BinaryCrossentropy()
loss2=tf.keras.losses.CategoricalCrossentropy()

# Compile the models
model1.compile(optimizer1, loss1)
model2.compile(optimizer2, loss2)

# Set tracking metrics
metricsToTrack = metrics.MeanSquaredError(name='MSE')

# Set early stopping condition and plateau monitor conditions
earlyStopping1 = EarlyStopping(monitor='val_loss', mode='auto', patience=5, restore_best_weights=True, min_delta=0.00000000001)
plateau1 = ReduceLROnPlateau(monitor='val_loss', mode='auto', patience=3, factor=0.01)

earlyStopping2 = EarlyStopping(monitor='val_loss', mode='auto', patience=5, restore_best_weights=True, min_delta=0.00000000001)
plateau2 = ReduceLROnPlateau(monitor='val_loss', mode='auto', patience=3, factor=0.01)

# Fit models and store results to plot network statisticsx
history1 = model1.fit(x=xTrain1, y=yTrain1, batch_size=1000, epochs=100, verbose=0, callbacks=[earlyStopping1, plateau1], validation_data=(xVerify1, yVerify1))
history2 = model2.fit(x=xTrain2, y=yTrain2, batch_size=1000, epochs=100, verbose=0, callbacks=[earlyStopping2, plateau2], validation_data=(xVerify2, yVerify2))

# Make predictions using the networks
prediction1 = model1.predict(xTest1)
prediction2 = model2.predict(xTest2)

predictedFlag1 = np.zeros(len(prediction1))
predictedFlag2 = np.zeros(len(prediction2))

fpr1, tpr1, thresholds1 = roc_curve(yTest1, prediction1)

gmeans1 = np.sqrt(tpr1 * (1-fpr1))

ix1 = np.argmax(gmeans1)

# Convert prediction percentages into flag values
for i in range(len(prediction1)):
    if prediction1[i] >= thresholds1[ix1]:
        predictedFlag1[i] = 1

for i in range(len(prediction2)):
    index = int(np.argmax(prediction2[i]))
    if index > 0:
        predictedFlag2[i] = index + 1
    else:
        predictedFlag2[i] = index
           
tv1 = 0
fv1 = 0
tv2 = 0
fv2 = 0

# Calculate true and false values (which predictions were correct)
for i in range(len(prediction1)):
    if yTest1.iloc[i]['flag'] == predictedFlag1[i]:
        tv1 += 1
    else:
        fv1 += 1
        
for i in range(len(prediction2)):
    if yTest2.iloc[i]['flag_0.0'] > 0 and predictedFlag2[i] == 0:
        tv2 += 1
    elif yTest2.iloc[i]['flag_2.0'] > 0 and predictedFlag2[i] == 2:
        tv2 += 1
    elif yTest2.iloc[i]['flag_3.0'] > 0 and predictedFlag2[i] == 3:
        tv2 += 1
    else:
        fv2 += 1
       
# Print true values and false value percentages
print('True Percentage: '+str(tv1 / len(prediction1)))
print('False Percentage: '+str(fv1 / len(prediction1)))
print('Loss: '+str(history1.history['val_loss'][-1]))

print('True Percentage: '+str(tv2 / len(prediction2)))
print('False Percentage: '+str(fv2 / len(prediction2)))
print('Loss: '+str(history2.history['val_loss'][-1]))

# TwoFlagsLoss.write('True Percentage: '+str(tv1 / len(prediction1))+'\n')
# TwoFlagsLoss.write('False Percentage: '+str(fv1 / len(prediction1))+'\n')
# TwoFlagsLoss.write('Loss: '+str(history1.history['val_loss'][-1])+'\n')

# ThreeFlagsLoss.write('True Percentage: '+str(tv2 / len(prediction2))+'\n')
# ThreeFlagsLoss.write('False Percentage: '+str(fv2 / len(prediction2))+'\n')
# ThreeFlagsLoss.write('Loss: '+str(history2.history['val_loss'][-1])+'\n')

# Save Models
# model1.save(outputFilePath+'mesaModelClassification2Flags.h5')
# model2.save(outputFilePath+'mesaModelClassification3Flags.h5')

plt.figure()
plt.plot(history1.history['loss'], label='Train')
plt.plot(history1.history['val_loss'], label='Validation')
plt.xlabel('epoch')
plt.ylabel('loss')
plt.title('Model Loss')
plt.legend()
# plt.savefig(outputFilePath+'MESAClassificationLoss2Flags.png')
plt.close()

plt.figure()
plt.plot(history2.history['loss'], label='Train')
plt.plot(history2.history['val_loss'], label='Validation')
plt.xlabel('epoch')
plt.ylabel('loss')
plt.title('Model Loss')
plt.legend()
# plt.savefig(outputFilePath+'MESAClassificationLoss3Flags.png')
plt.close()
