#!/usr/bin/python3.8
"""
Date Created:
Date Edited: Tuesday 31 May 2022
Author: Mitchell T Dennis

A script to create neural networks which predict output data. The input data is of the form Mass, Helium Abundance, and Metallicity. The two networks are learning for TRGB atmospheric parameters (log(Luminosity) in LSun units, log(Surface Gravity), Effective Temperature, and [Fe/H] and IBand Magnitude (Johnsons-Cousins) and associated Empirical error from the Worthey and Lee Bolometric correction code (Worthey and Lee 2011).
"""

import tensorflow as tf
import numpy as np
import pandas as pd
import datetime
from pathlib import Path
import sys
from sklearn import preprocessing
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt
from tensorflow.keras import metrics
from tensorflow.keras.callbacks import EarlyStopping, LearningRateScheduler, ReduceLROnPlateau
from machineLearningFunctions import *
import pickle as pkl

# Set initial constants
hiddenNodes = 24
scale = 1
layers = 3
dropout = 0
numRuns = int(sys.argv[1])

now = datetime.datetime.now()
dt_string = now.strftime("%d-%m-%y_%H-%M-%S")
home = str(Path.home())

# Open data files
outputFilePath = '/home/mtde226/MESA_emulator/LOGS/LOGS_'+str(numRuns)+'/MachineLearningResults/'
inputFilePath = '/home/mtde226/MESA_emulator/LOGS/LOGS_'+str(numRuns)+'/postProcessResults/fullComboData.txt'
normFile = open(outputFilePath+'normalisationConstantsCombo.txt', 'r')
TRGBLoss = open(outputFilePath+'TRGBLossStatistics.txt', 'w')
IBandLoss = open(outputFilePath+'IBandLossStatistics.txt', 'w')

TRGBErrorDistTeff = open(outputFilePath+'teffError.pkl', 'wb')
TRGBErrorDistLum = open(outputFilePath+'lumError.pkl', 'wb')
TRGBErrorDistGrav = open(outputFilePath+'gravError.pkl', 'wb')
TRGBErrorDistFeH = open(outputFilePath+'fehError.pkl', 'wb')
IBandErrorDistIBand = open(outputFilePath+'ibandError.pkl', 'wb')
IBandErrorDistIerr = open(outputFilePath+'ierrError.pkl', 'wb')
IBandErrorDistVIBand = open(outputFilePath+'vibandError.pkl', 'wb')
IBandErrorDistVIerr = open(outputFilePath+'vierrError.pkl', 'wb')

# Create dataframes and clean data
data = pd.read_csv(inputFilePath, header=0, index_col=False)
inputCols = ['mass', 'Y', 'Z', 'A']
outputCols1 = ['grav', 'Teff', 'FeH', 'logL']
outputCols2 = ['IBand', 'Ierr','VI','VIerr']
normalCols = ['mass','Y','Z','A','time','grav','Teff','FeH','logL','IBand','Ierr','VI','VIerr']

dataCols1 = inputCols + outputCols1
dataCols2 = inputCols + outputCols2

# print(len(data.query('flag == 2')))
# print(len(data.query('IBand < -4.2')))
# print(len(data.query('flag == 2 and IBand > -4.2')))

# Read the normalisation constants
massABSMin, massMin, massMax = [float(x) for x in normFile.readline().split()]
heliumABSMin, heliumMin, heliumMax = [float(x) for x in normFile.readline().split()]
metallicityABSMin, metallicityMin, metallicityMax = [float(x) for x in normFile.readline().split()]
axionABSMin, axionMin, axionMax = [float(x) for x in normFile.readline().split()]
timeABSMin, timeMin, timeMax = [float(x) for x in normFile.readline().split()]
gravABSMin, gravMin, gravMax = [float(x) for x in normFile.readline().split()]
TeffABSMin, TeffMin, TeffMax = [float(x) for x in normFile.readline().split()]
FeHABSMin, FeHMin, FeHMax = [float(x) for x in normFile.readline().split()]
logLABSMin, logLMin, logLMax = [float(x) for x in normFile.readline().split()]
IBandABSMin, IBandMin, IBandMax = [float(x) for x in normFile.readline().split()]
IerrABSMin, IerrMin, IerrMax = [float(x) for x in normFile.readline().split()]
VIBandABSMin, VIBandMin, VIBandMax = [float(x) for x in normFile.readline().split()]
VIerrABSMin, VIerrMin, VIerrMax = [float(x) for x in normFile.readline().split()]

exclusionIBand = normaliseFromValues(IBandMin, IBandMax, IBandABSMin, -2.5)
exclusionTeff = normaliseFromValues(TeffMin, TeffMax, TeffABSMin, 8000)

# Exclude certain data
data.query('flag == 0', inplace=True)
data.query('Teff < @exclusionTeff', inplace=True)

step=4

"""
dropNumsX = range(0,10,step)
dropNumsY = range(0,15,step)
dropNumsZ = range(0,25,step)
dropNumsA = range(0,30,step)

oldData = len(data)

print(data.columns)
for i in dropNumsX:
    data.query('i != @i', inplace=True)

for j in dropNumsY:
    data.query('j != @j', inplace=True)

for k in dropNumsZ:
    data.query('k != @k', inplace=True)

for l in dropNumsA:
    data.query('l != @l', inplace=True)
    
print(len(data)/oldData)
"""

# Always use deep copies of data to avoid interference between data sets
data1 = data[dataCols1].copy(deep=True)
data2 = data[dataCols2].copy(deep=True)

# Create training and testing data
xTrain1, yTrain1, xTest1, yTest1, xVerify1, yVerify1 = splitData(data1, inputCols, outputCols1, 0.80, seed=1)
xTrain2, yTrain2, xTest2, yTest2, xVerify2, yVerify2 = splitData(data2, inputCols, outputCols2, 0.80, seed=1)

# Create the two models; one for TRGB atmospheric data and the other for IBand data
model1=createModel(inputDim=len(inputCols), outputDim=len(outputCols1), numLayers=layers, hiddenInit=hiddenNodes, dropoutFrac=dropout, scaling=scale)
model2=createModel(inputDim=len(inputCols), outputDim=len(outputCols2), numLayers=layers, hiddenInit=hiddenNodes, dropoutFrac=dropout, scaling=scale)

# IMPORTANT: Each network needs an independent optimizers, early stopping, plateau monitors, etc...
# Set the optimizers
optimizer1 = tf.keras.optimizers.Adam(learning_rate=0.04, beta_1=0.15, beta_2=0.4, amsgrad=False)
optimizer2 = tf.keras.optimizers.Adam(learning_rate=0.04, beta_1=0.15, beta_2=0.4, amsgrad=False)

# Set the loss functions
loss1=tf.keras.losses.MeanSquaredError()
loss2=tf.keras.losses.MeanSquaredError()

# Compile the models
model1.compile(optimizer1, loss1)
model2.compile(optimizer2, loss2)

# Set tracking metrics
metricsToTrack = metrics.MeanSquaredError(name='MSE')

# Set early stopping condition and plateau monitor conditions
earlyStopping1 = EarlyStopping(monitor='val_loss', mode='auto', patience=10, restore_best_weights=True, min_delta=0.00000002)
plateau1 = ReduceLROnPlateau(monitor='val_loss', mode='auto', patience=3, factor=0.1, min_lr=0.0000000000000001)

earlyStopping2 = EarlyStopping(monitor='val_loss', mode='auto', patience=10, restore_best_weights=True, min_delta=0.00000002)
plateau2 = ReduceLROnPlateau(monitor='val_loss', mode='auto', patience=3, factor=0.1, min_lr=0.000000000000001)

# Fit models and store results to plot network statistics
history1 = model1.fit(x=xTrain1, y=yTrain1, batch_size=500, epochs=100, verbose=0, callbacks=[earlyStopping1, plateau1], validation_data=(xVerify1, yVerify1))
history2 = model2.fit(x=xTrain2, y=yTrain2, batch_size=500, epochs=100, verbose=0, callbacks=[earlyStopping2, plateau2], validation_data=(xVerify2, yVerify2))

# Make predictions using the networks
prediction1 = model1.predict(xTest1)
prediction2 = model2.predict(xTest2)

# Denormalise data and prediction results
yTest1['Teff'] =  deNormalise(TeffMin, TeffMax, TeffABSMin, yTest1['Teff'])
yTest1['grav'] =  deNormalise(gravMin, gravMax, gravABSMin, yTest1['grav'])
yTest1['FeH'] =  deNormalise(FeHMin, FeHMax, FeHABSMin, yTest1['FeH'])
yTest1['logL'] =  deNormalise(logLMin, logLMax, logLABSMin, yTest1['logL'])
yTest2['IBand'] =  deNormalise(IBandMin, IBandMax, IBandABSMin, yTest2['IBand'])
yTest2['Ierr'] =  deNormalise(IerrMin, IerrMax, IerrABSMin, yTest2['Ierr'])
yTest2['VI'] =  deNormalise(VIBandMin, VIBandMax, VIBandABSMin, yTest2['VI'])
yTest2['VIerr'] =  deNormalise(VIerrMin, VIerrMax, VIerrABSMin, yTest2['VIerr'])

prediction1[:, 0] = deNormalise(gravMin, gravMax, gravABSMin, prediction1[:, 0])
prediction1[:, 1] = deNormalise(TeffMin, TeffMax, TeffABSMin, prediction1[:, 1])
prediction1[:, 2] = deNormalise(FeHMin, FeHMax, FeHABSMin, prediction1[:, 2])
prediction1[:, 3] = deNormalise(logLMin, logLMax, logLABSMin, prediction1[:, 3])
prediction2[:, 0] = deNormalise(IBandMin, IBandMax, IBandABSMin, prediction2[:, 0])
prediction2[:, 1] = deNormalise(IerrMin, IerrMax, IerrABSMin, prediction2[:, 1])
prediction2[:, 2] = deNormalise(VIBandMin, VIBandMax, VIBandABSMin, prediction2[:, 2])
prediction2[:, 3] = deNormalise(VIerrMin, VIerrMax, VIerrABSMin, prediction2[:, 3])

print(mean_squared_error(yTest1['grav'], prediction1[:, 0], squared=False))
print(mean_squared_error(yTest1['Teff'], prediction1[:, 1], squared=False))
print(mean_squared_error(yTest1['FeH'], prediction1[:, 2], squared=False))
print(mean_squared_error(yTest1['logL'], prediction1[:, 3], squared=False))
print('Loss: '+str(history1.history['val_loss'][-1]))
print(mean_squared_error(yTest2['IBand'], prediction2[:, 0], squared=False))
print(mean_squared_error(yTest2['Ierr'], prediction2[:, 1], squared=False))
print(mean_squared_error(yTest2['VI'], prediction2[:, 2], squared=False))
print(mean_squared_error(yTest2['VIerr'], prediction2[:, 3], squared=False))
print('Loss: '+str(history2.history['val_loss'][-1]))

TRGBLoss.write('grav RMSE: '+str(mean_squared_error(yTest1['grav'], prediction1[:, 0], squared=False))+'\n')
TRGBLoss.write('Teff RMSE: '+str(mean_squared_error(yTest1['Teff'], prediction1[:, 1], squared=False))+'\n')
TRGBLoss.write('FeH RMSE: '+str(mean_squared_error(yTest1['FeH'], prediction1[:, 2], squared=False))+'\n')
TRGBLoss.write('logL RMSE: '+str(mean_squared_error(yTest1['logL'], prediction1[:, 3], squared=False))+'\n')
TRGBLoss.write('Loss: '+str(history1.history['val_loss'][-1])+'\n')

IBandLoss.write('IBand RMSE: '+str(mean_squared_error(yTest2['IBand'], prediction2[:, 0], squared=False))+'\n')
IBandLoss.write('Ierr RMSE: '+str(mean_squared_error(yTest2['Ierr'], prediction2[:, 1], squared=False))+'\n')
IBandLoss.write('IBand RMSE: '+str(mean_squared_error(yTest2['VI'], prediction2[:, 2], squared=False))+'\n')
IBandLoss.write('Ierr RMSE: '+str(mean_squared_error(yTest2['VIerr'], prediction2[:, 3], squared=False))+'\n')
IBandLoss.write('Loss: '+str(history2.history['val_loss'][-1])+'\n')

pkl.dump(yTest1['grav'].to_numpy() - prediction1[:, 0], TRGBErrorDistGrav)
pkl.dump(yTest1['Teff'].to_numpy() - prediction1[:, 1], TRGBErrorDistTeff)
pkl.dump(yTest1['FeH'].to_numpy() - prediction1[:, 2], TRGBErrorDistFeH)
pkl.dump(yTest1['logL'].to_numpy() - prediction1[:, 3], TRGBErrorDistLum)

pkl.dump(yTest2['IBand'].to_numpy() - prediction2[:, 0], IBandErrorDistIBand)
pkl.dump(yTest2['Ierr'].to_numpy() - prediction2[:, 1], IBandErrorDistIerr)
pkl.dump(yTest2['VI'].to_numpy() - prediction2[:, 2], IBandErrorDistVIBand)
pkl.dump(yTest2['VIerr'].to_numpy() - prediction2[:, 3], IBandErrorDistVIerr)

# Save models
model1.save(outputFilePath+'mesaModelTRGB.h5')
model2.save(outputFilePath+'mesaModelIBand.h5')

# Plot useful figures
plt.figure()
plt.plot(history1.history['loss'], label='Train')
plt.plot(history1.history['val_loss'], label='Validation')
plt.xlabel('epoch')
plt.ylabel('loss')
plt.yscale('log')
plt.title('Model Loss')
plt.legend()
plt.savefig(outputFilePath+'TRGBLoss.png')
plt.close()

plt.figure()
plt.plot(history2.history['loss'], label='Train')
plt.plot(history2.history['val_loss'], label='Validation')
plt.xlabel('epoch')
plt.ylabel('loss')
plt.yscale('log')
plt.title('Model Loss')
plt.legend()
plt.savefig(outputFilePath+'IBandLoss.png')
plt.close()

plt.figure()
plt.hist2d(yTest1['grav'].values, prediction1[:, 0], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('Surface Gravity')
plt.colorbar()
plt.savefig(outputFilePath+'SurfaceG.png')
plt.close()

plt.figure()
plt.hist2d(yTest1['FeH'].values, prediction1[:, 2], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('[Fe/H]')
plt.colorbar()
plt.savefig(outputFilePath+'FeH.png')
plt.close()

plt.figure()
plt.hist2d(yTest1['logL'].values, prediction1[:, 3], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('Luminosity')
plt.colorbar()
plt.savefig(outputFilePath+'Luminosity.png')
plt.close()

plt.figure()
plt.hist2d(yTest1['Teff'].values, prediction1[:, 1], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('Effective Temperature')
plt.colorbar()
plt.savefig(outputFilePath+'EffectiveT.png')
plt.close()

plt.figure()
plt.hist2d(yTest2['IBand'].values, prediction2[:, 0], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('I Band Magnitude')
plt.colorbar()
plt.savefig(outputFilePath+'IBand.png')
plt.close()

plt.figure()
plt.hist2d(yTest2['Ierr'].values, prediction2[:, 1], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('I Band Error')
plt.colorbar()
plt.savefig(outputFilePath+'Ierr.png')
plt.close()

plt.figure()
plt.hist2d(yTest2['VI'].values, prediction2[:, 2], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('I Band Magnitude')
plt.colorbar()
plt.savefig(outputFilePath+'VI.png')
plt.close()

plt.figure()
plt.hist2d(yTest2['VIerr'].values, prediction2[:, 3], bins=[20, 20])
plt.xlabel('True Value')
plt.ylabel('Predicted Value')
plt.title('I Band Error')
plt.colorbar()
plt.savefig(outputFilePath+'VIerr.png')
plt.close()
